package com.autentia.tutoriales.spring.aop.aspectj.cache;

import java.lang.reflect.Method;
import com.opensymphony.oscache.general.GeneralCacheAdministrator;
import com.opensymphony.oscache.web.filter.ExpiresRefreshPolicy;
import com.opensymphony.oscache.base.NeedsRefreshException;
import java.util.Properties;
import org.aspectj.lang.Signature;

/**
 * Aspecto que se encarga del sistema de cache de la aplicación
 * @author Carlos García. Autentia.
 * @see http://www.mobiletest.es
 */
@org.aspectj.lang.annotation.Aspect
public class CacheAspect {
	
	private final int	 NO_EXPIRE	 = 0;
	private final String CACHE_KEY	 = "cacheKey";
	private final String EXPIRE_TIME = "expireTime";
	
	private GeneralCacheAdministrator cache;
	
	/**
	 * Sí, si.. está acoplado a OSCache, podría haber definido una inferfaz que me desacoplara del sistema de cache... 
	 * Recuerda que esto es un tutorial y no tenia ganas.
	 * @param cache Sistema de cache nativo, no reinventemos la rueda :-) 
	 */
	public CacheAspect(com.opensymphony.oscache.general.GeneralCacheAdministrator cache){
		this.cache = cache;
	}
	
	/**
	 * Este método será llamado por AOP en tiempo de ejecución para aquellos 
	 * métodos con la anotación Cachea
	 */
	@org.aspectj.lang.annotation.Around("@annotation(Cachea)")
	public Object cachear(org.aspectj.lang.ProceedingJoinPoint call) throws Throwable {
		Properties cacheProps = this.getAnnotationProperties(call, true);
	    Object	 result		  = null;
	    String	 cacheKey	  = cacheProps.getProperty(CACHE_KEY);
	    int		 expire		  = NO_EXPIRE;

	    expire	 = Integer.parseInt(cacheProps.getProperty(EXPIRE_TIME));
    	// Evitamos números negativos como valor de tiempo de validez de la cache
    	if (expire < 0){
    		expire = NO_EXPIRE;
    	}
	    
	    try {  
	    	 result  = cache.getFromCache(cacheKey);
	    } catch (NeedsRefreshException e) {
	    	 // Los datos de la cache no existen o han caducado
	    	cache.cancelUpdate(cacheKey);  
	    }  	    
	   
	    if (result == null) {
	    	// Actualmente no hay datos cacheados validos, ejecutamos el método y 
	    	// cacheamos el resultado.  
	    	result = call.proceed();
	    	
	    	if (expire == 0){
	    		// No se especificó un tiempo de validez
	    		cache.putInCache(cacheKey, result);	
	    	} else {
	    		// Tiene un tiempo de validez
	    		cache.putInCache(cacheKey, result, new ExpiresRefreshPolicy(expire));
	    	}
		}
		
		return result;
	}
	
	/**
	 * Este método será llamado por AOP en tiempo de ejecución.
	 * Vemos si el método tiene parámetros de descacheo a través de la anotación Descachea 
	 */	
	@org.aspectj.lang.annotation.Around("@annotation(Descachea)")
	public Object descachear(org.aspectj.lang.ProceedingJoinPoint call) throws Throwable {
		// Ejecutamos el método
		Object result	 = call.proceed();
		
		// Descacheamos
		Properties cacheProps = this.getAnnotationProperties(call, false);
		String	  cacheKey	  = cacheProps.getProperty(CACHE_KEY);		
		cache.removeEntry(cacheKey);

		// Devolvemos el resultado del método
		return result;
	}	

	
	/**
	 * @return Devuelve un properties con los atributos de la anotación (Cachea o Descachea)
	 */	
	private Properties getAnnotationProperties(org.aspectj.lang.ProceedingJoinPoint call, boolean isCacheo) {
		Properties	properties = new Properties();
		Method		metodo	   = this.getCallMethod(call);
		
		if (isCacheo){
			Cachea	anotacion  = metodo.getAnnotation(Cachea.class);
			properties.put(CACHE_KEY, anotacion.cacheKey());
			properties.put(EXPIRE_TIME, String.valueOf(anotacion.expireTime()));
		} else {
			Descachea anotacion = metodo.getAnnotation(Descachea.class);
			properties.put(CACHE_KEY, anotacion.cacheKey());
		}
		return properties;
	}

	
	/**
	 * @return Devuelve una referencia al método invocado por AOP
	 */
	@SuppressWarnings("unchecked")
	private Method getCallMethod(org.aspectj.lang.ProceedingJoinPoint call){
		Method metodo	  =  null;
		
		try {
			Signature sig     = call.getSignature();
			Class  clase	  = sig.getDeclaringType();
			String methodName = sig.getName();
			Object[] args	  = call.getArgs();
			Class[] params	  = new Class[args.length];
			
			for (int i = 0, count = args.length; i < count; i++){
				params[i] = args[i].getClass();
			}
			
			metodo	 = clase.getMethod(methodName, params);
		} catch (Exception e) {
			// Ignoramos: SecurityException, NoSuchMethodException
		}
		
		return metodo;
	}	
}
